@@ -17,7 +17,15 @@
 # include <sys/ipc.h>
 # include <sys/msg.h>
 # include <sys/sem.h>
-#endif /* ! FAKEROOT_FAKENET */
+#else /* FAKEROOT_FAKENET */
+# include <netinet/in.h>
+# include <netinet/tcp.h>
+# include <netdb.h>
+# include <pthread.h>
+# ifdef HAVE_ENDIAN_H
+#  include <endian.h>
+# endif
+#endif /* FAKEROOT_FAKENET */
 #include <fcntl.h>
 #include <unistd.h>
 #include <errno.h>
@@ -29,20 +37,9 @@
 # include <sys/socket.h>
 #endif
 
-#ifndef FAKEROOT_FAKENET
-int msg_snd=-1;
-int msg_get=-1;
-int sem_id=-1;
-#else /* FAKEROOT_FAKENET */
-#include <netinet/in.h>
-#include <netinet/tcp.h>
-#include <netdb.h>
-#include <pthread.h>
-#ifdef HAVE_ENDIAN_H
-# include <endian.h>
+#ifdef STUPID_ALPHA_HACK
+#include "stats.h"
 #endif
-#endif /* FAKEROOT_FAKENET */
-
 
 #ifndef _UTSNAME_LENGTH
 /* for LINUX libc5 */
@@ -50,11 +47,17 @@ int sem_id=-1;
 #endif
 
 
-#ifdef FAKEROOT_FAKENET
-static int comm_sd = -1;
+#ifndef FAKEROOT_FAKENET
+int msg_snd=-1;
+int msg_get=-1;
+int sem_id=-1;
+#else /* FAKEROOT_FAKENET */
+volatile int comm_sd = -1;
 static pthread_mutex_t comm_sd_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif /* FAKEROOT_FAKENET */
 
 
+#ifdef FAKEROOT_FAKENET
 static void fail(const char *msg)
 {
   if (errno > 0)
@@ -64,9 +67,8 @@ static void fail(const char *msg)
 
   exit(1);
 }
-
-
 #endif /* FAKEROOT_FAKENET */
+
 const char *env_var_set(const char *env){
   const char *s;
   
@@ -78,9 +80,13 @@ const char *env_var_set(const char *env){
     return NULL;
 }
 
-void cpyfakemstat(struct fake_msg *f,
-                 const struct stat     *st){
+void cpyfakemstat(struct fake_msg *f, const struct stat *st
+#ifdef STUPID_ALPHA_HACK
+		, int ver
+#endif
+		){
   
+#ifndef STUPID_ALPHA_HACK
   f->st.mode =st->st_mode;
   f->st.ino  =st->st_ino ;
   f->st.uid  =st->st_uid ;
@@ -97,10 +103,54 @@ void cpyfakemstat(struct fake_msg *f,
      to occur in practical fakeroot conditions. */
 
   f->st.nlink=st->st_nlink;
+#else
+  switch(ver) {
+	  case _STAT_VER_KERNEL:
+  f->st.mode  = ((struct fakeroot_kernel_stat *)st)->st_mode;
+  f->st.ino   = ((struct fakeroot_kernel_stat *)st)->st_ino;
+  f->st.uid   = ((struct fakeroot_kernel_stat *)st)->st_uid;
+  f->st.gid   = ((struct fakeroot_kernel_stat *)st)->st_gid;
+  f->st.dev   = ((struct fakeroot_kernel_stat *)st)->st_dev;
+  f->st.rdev  = ((struct fakeroot_kernel_stat *)st)->st_rdev;
+  f->st.nlink = ((struct fakeroot_kernel_stat *)st)->st_nlink;
+  break;
+	  case _STAT_VER_GLIBC2:
+  f->st.mode  = ((struct fakeroot_glibc2_stat *)st)->st_mode;
+  f->st.ino   = ((struct fakeroot_glibc2_stat *)st)->st_ino;
+  f->st.uid   = ((struct fakeroot_glibc2_stat *)st)->st_uid;
+  f->st.gid   = ((struct fakeroot_glibc2_stat *)st)->st_gid;
+  f->st.dev   = ((struct fakeroot_glibc2_stat *)st)->st_dev;
+  f->st.rdev  = ((struct fakeroot_glibc2_stat *)st)->st_rdev;
+  f->st.nlink = ((struct fakeroot_glibc2_stat *)st)->st_nlink;
+  break;
+		  case _STAT_VER_GLIBC2_1:
+  f->st.mode  = ((struct fakeroot_glibc21_stat *)st)->st_mode;
+  f->st.ino   = ((struct fakeroot_glibc21_stat *)st)->st_ino;
+  f->st.uid   = ((struct fakeroot_glibc21_stat *)st)->st_uid;
+  f->st.gid   = ((struct fakeroot_glibc21_stat *)st)->st_gid;
+  f->st.dev   = ((struct fakeroot_glibc21_stat *)st)->st_dev;
+  f->st.rdev  = ((struct fakeroot_glibc21_stat *)st)->st_rdev;
+  f->st.nlink = ((struct fakeroot_glibc21_stat *)st)->st_nlink;
+  break;
+		  default:
+  f->st.mode  = st->st_mode;
+  f->st.ino   = st->st_ino;
+  f->st.uid   = st->st_uid;
+  f->st.gid   = st->st_gid;
+  f->st.dev   = st->st_dev;
+  f->st.rdev  = st->st_rdev;
+  f->st.nlink = st->st_nlink;
+  break;
+  }
+#endif
 }
 
-void cpystatfakem(struct     stat *st,
-                 const struct fake_msg *f){
+void cpystatfakem(struct stat *st, const struct fake_msg *f
+#ifdef STUPID_ALPHA_HACK
+		, int ver
+#endif
+		){
+#ifndef STUPID_ALPHA_HACK
   st->st_mode =f->st.mode;
   st->st_ino  =f->st.ino ;
   st->st_uid  =f->st.uid ;
@@ -110,13 +160,53 @@ void cpystatfakem(struct     stat *st,
   /* DON'T copy the nlink count! The system always knows
      this one better! */
   /*  st->st_nlink=f->st.nlink;*/
+#else
+  switch(ver) {
+	  case _STAT_VER_KERNEL:
+  ((struct fakeroot_kernel_stat *)st)->st_mode = f->st.mode;
+  ((struct fakeroot_kernel_stat *)st)->st_ino  = f->st.ino;
+  ((struct fakeroot_kernel_stat *)st)->st_uid  = f->st.uid;
+  ((struct fakeroot_kernel_stat *)st)->st_gid  = f->st.gid;
+  ((struct fakeroot_kernel_stat *)st)->st_dev  = f->st.dev;
+  ((struct fakeroot_kernel_stat *)st)->st_rdev = f->st.rdev;
+  break;
+	  case _STAT_VER_GLIBC2:
+  ((struct fakeroot_glibc2_stat *)st)->st_mode = f->st.mode;
+  ((struct fakeroot_glibc2_stat *)st)->st_ino  = f->st.ino;
+  ((struct fakeroot_glibc2_stat *)st)->st_uid  = f->st.uid;
+  ((struct fakeroot_glibc2_stat *)st)->st_gid  = f->st.gid;
+  ((struct fakeroot_glibc2_stat *)st)->st_dev  = f->st.dev;
+  ((struct fakeroot_glibc2_stat *)st)->st_rdev = f->st.rdev;
+  break;
+		  case _STAT_VER_GLIBC2_1:
+  ((struct fakeroot_glibc21_stat *)st)->st_mode = f->st.mode;
+  ((struct fakeroot_glibc21_stat *)st)->st_ino  = f->st.ino;
+  ((struct fakeroot_glibc21_stat *)st)->st_uid  = f->st.uid;
+  ((struct fakeroot_glibc21_stat *)st)->st_gid  = f->st.gid;
+  ((struct fakeroot_glibc21_stat *)st)->st_dev  = f->st.dev;
+  ((struct fakeroot_glibc21_stat *)st)->st_rdev = f->st.rdev;
+  break;
+		  default:
+  st->st_mode =f->st.mode;
+  st->st_ino  =f->st.ino ;
+  st->st_uid  =f->st.uid ;
+  st->st_gid  =f->st.gid ;
+  st->st_dev  =f->st.dev ;
+  st->st_rdev =f->st.rdev;
+  break;
+  }
+#endif
 }
 
 #ifdef STAT64_SUPPORT
 
 void cpyfakemstat64(struct fake_msg *f,
-                 const struct stat64 *st){
-  
+                 const struct stat64 *st
+#ifdef STUPID_ALPHA_HACK
+                 , int ver
+#endif
+                 ){
+#ifndef STUPID_ALPHA_HACK
   f->st.mode =st->st_mode;
   f->st.ino  =st->st_ino ;
   f->st.uid  =st->st_uid ;
@@ -133,9 +223,54 @@ void cpyfakemstat64(struct fake_msg *f,
      to occur in practical fakeroot conditions. */
 
   f->st.nlink=st->st_nlink;
+#else
+  switch(ver) {
+	  case _STAT_VER_KERNEL:
+  f->st.mode  = ((struct fakeroot_kernel_stat *)st)->st_mode;
+  f->st.ino   = ((struct fakeroot_kernel_stat *)st)->st_ino;
+  f->st.uid   = ((struct fakeroot_kernel_stat *)st)->st_uid;
+  f->st.gid   = ((struct fakeroot_kernel_stat *)st)->st_gid;
+  f->st.dev   = ((struct fakeroot_kernel_stat *)st)->st_dev;
+  f->st.rdev  = ((struct fakeroot_kernel_stat *)st)->st_rdev;
+  f->st.nlink = ((struct fakeroot_kernel_stat *)st)->st_nlink;
+  break;
+	  case _STAT_VER_GLIBC2:
+  f->st.mode  = ((struct fakeroot_glibc2_stat *)st)->st_mode;
+  f->st.ino   = ((struct fakeroot_glibc2_stat *)st)->st_ino;
+  f->st.uid   = ((struct fakeroot_glibc2_stat *)st)->st_uid;
+  f->st.gid   = ((struct fakeroot_glibc2_stat *)st)->st_gid;
+  f->st.dev   = ((struct fakeroot_glibc2_stat *)st)->st_dev;
+  f->st.rdev  = ((struct fakeroot_glibc2_stat *)st)->st_rdev;
+  f->st.nlink = ((struct fakeroot_glibc2_stat *)st)->st_nlink;
+  break;
+	  case _STAT_VER_GLIBC2_1:
+  f->st.mode  = ((struct fakeroot_glibc21_stat *)st)->st_mode;
+  f->st.ino   = ((struct fakeroot_glibc21_stat *)st)->st_ino;
+  f->st.uid   = ((struct fakeroot_glibc21_stat *)st)->st_uid;
+  f->st.gid   = ((struct fakeroot_glibc21_stat *)st)->st_gid;
+  f->st.dev   = ((struct fakeroot_glibc21_stat *)st)->st_dev;
+  f->st.rdev  = ((struct fakeroot_glibc21_stat *)st)->st_rdev;
+  f->st.nlink = ((struct fakeroot_glibc21_stat *)st)->st_nlink;
+  break;
+		  default:
+  f->st.mode  = st->st_mode;
+  f->st.ino   = st->st_ino;
+  f->st.uid   = st->st_uid;
+  f->st.gid   = st->st_gid;
+  f->st.dev   = st->st_dev;
+  f->st.rdev  = st->st_rdev;
+  f->st.nlink = st->st_nlink;
+  break;
+  }
+#endif
 }
 void cpystat64fakem(struct stat64 *st,
-                 const struct fake_msg *f){
+                 const struct fake_msg *f
+#ifdef STUPID_ALPHA_HACK
+                 , int ver
+#endif
+                 ){
+#ifndef STUPID_ALPHA_HACK
   st->st_mode =f->st.mode;
   st->st_ino  =f->st.ino ;
   st->st_uid  =f->st.uid ;
@@ -145,6 +280,42 @@ void cpystat64fakem(struct stat64 *st,
   /* DON'T copy the nlink count! The system always knows
      this one better! */
   /*  st->st_nlink=f->st.nlink;*/
+#else
+  switch(ver) {
+	  case _STAT_VER_KERNEL:
+  ((struct fakeroot_kernel_stat *)st)->st_mode = f->st.mode;
+  ((struct fakeroot_kernel_stat *)st)->st_ino  = f->st.ino;
+  ((struct fakeroot_kernel_stat *)st)->st_uid  = f->st.uid;
+  ((struct fakeroot_kernel_stat *)st)->st_gid  = f->st.gid;
+  ((struct fakeroot_kernel_stat *)st)->st_dev  = f->st.dev;
+  ((struct fakeroot_kernel_stat *)st)->st_rdev = f->st.rdev;
+  break;
+	  case _STAT_VER_GLIBC2:
+  ((struct fakeroot_glibc2_stat *)st)->st_mode = f->st.mode;
+  ((struct fakeroot_glibc2_stat *)st)->st_ino  = f->st.ino;
+  ((struct fakeroot_glibc2_stat *)st)->st_uid  = f->st.uid;
+  ((struct fakeroot_glibc2_stat *)st)->st_gid  = f->st.gid;
+  ((struct fakeroot_glibc2_stat *)st)->st_dev  = f->st.dev;
+  ((struct fakeroot_glibc2_stat *)st)->st_rdev = f->st.rdev;
+  break;
+		  case _STAT_VER_GLIBC2_1:
+  ((struct fakeroot_glibc21_stat *)st)->st_mode = f->st.mode;
+  ((struct fakeroot_glibc21_stat *)st)->st_ino  = f->st.ino;
+  ((struct fakeroot_glibc21_stat *)st)->st_uid  = f->st.uid;
+  ((struct fakeroot_glibc21_stat *)st)->st_gid  = f->st.gid;
+  ((struct fakeroot_glibc21_stat *)st)->st_dev  = f->st.dev;
+  ((struct fakeroot_glibc21_stat *)st)->st_rdev = f->st.rdev;
+  break;
+		  default:
+  st->st_mode =f->st.mode;
+  st->st_ino  =f->st.ino ;
+  st->st_uid  =f->st.uid ;
+  st->st_gid  =f->st.gid ;
+  st->st_dev  =f->st.dev ;
+  st->st_rdev =f->st.rdev;
+  break;
+  }
+#endif
 }
 
 #endif /* STAT64_SUPPORT */
@@ -206,6 +377,7 @@ void stat32from64(struct stat *s32, const struct stat64 *s64)
 #endif
 
 #ifndef FAKEROOT_FAKENET
+
 void semaphore_up(){
   struct sembuf op;
   if(sem_id==-1)
@@ -223,7 +395,30 @@ void semaphore_up(){
     } else {
       break;
     }
+  }
+}
+
+void semaphore_down(){
+  struct sembuf op;
+  if(sem_id==-1)
+    sem_id=semget(get_ipc_key()+2,1,IPC_CREAT|0600);
+  op.sem_num=0;
+  op.sem_op=1;
+  op.sem_flg=SEM_UNDO;
+  while (1) {
+    if (semop(sem_id,&op,1)) {
+      if (errno != EINTR) {
+        perror("semop(2): encountered an error");
+        exit(1);
+      }
+    } else {
+      break;
+    }
+  }
+}
+
 #else /* FAKEROOT_FAKENET */
+
 static struct sockaddr *get_addr(void)
 {
   static struct sockaddr_in addr = { 0, 0, { 0 } };
@@ -237,8 +432,6 @@ static struct sockaddr *get_addr(void)
       errno = 0;
       fail("FAKEROOTKEY not defined in environment");
     }
-#endif /* FAKEROOT_FAKENET */
-#ifdef FAKEROOT_FAKENET
 
     port = atoi(str);
     if (port <= 0 || port >= 65536) {
@@ -249,114 +442,141 @@ static struct sockaddr *get_addr(void)
     addr.sin_family = AF_INET;
     addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
     addr.sin_port = htons(port);
-#endif /* FAKEROOT_FAKENET */
   }
-#ifdef FAKEROOT_FAKENET
 
   return (struct sockaddr *) &addr;
-#endif /* FAKEROOT_FAKENET */
 }
-#ifndef FAKEROOT_FAKENET
-void semaphore_down(){
-  struct sembuf op;
-  if(sem_id==-1)
-    sem_id=semget(get_ipc_key()+2,1,IPC_CREAT|0600);
-  op.sem_num=0;
-  op.sem_op=1;
-  op.sem_flg=SEM_UNDO;
-  while (1) {
-    if (semop(sem_id,&op,1)) {
-      if (errno != EINTR) {
-	perror("semop(2): encountered an error");
-        exit(1);
-#else /* FAKEROOT_FAKENET */
 
 static void open_comm_sd(void)
 {
-  char *str;
-  int base, val;
-
-  if (comm_sd >= 0) {
-    int rc;
-    socklen_t len;
-
-    val = 0;
-    len = sizeof (val);
-    rc = getsockopt(comm_sd, SOL_SOCKET, SO_TYPE, &val, &len);
-
-    if (rc >= 0 && val == SOCK_STREAM) {
-      val = 0;
-      len = sizeof (val);
-      rc = getsockopt(comm_sd, SOL_SOCKET, SO_REUSEADDR, &val, &len);
-
-      if (rc >= 0 && val)
-        return;
-    }
-
-    fprintf(stderr, "libfakeroot: file descriptor %d was hijacked ", comm_sd);
-    if (rc < 0)
-      fprintf(stderr, "(getsockopt said \"%s\")\n", strerror(errno));
-    else
-      fprintf(stderr, "(socket options changed)\n");
-
-    comm_sd = -1;
-  }
-
-  base = getdtablesize() - 100;
-  if (base < 3)
-    base = 3;
-
-  if ((str = getenv(FD_BASE_ENV))) {
-    int fd = atoi(str);
-    if (fd >= 3)
-      base = fd;
-  }
+  if (comm_sd >= 0)
+    return;
 
   comm_sd = socket(PF_INET, SOCK_STREAM, 0);
   if (comm_sd < 0)
     fail("socket");
 
-  if (comm_sd < base) {
-    int fd;
-
-    do {
-      fd = fcntl(comm_sd, F_DUPFD, base);
-      if (fd < 0) {
-	if (errno == EINTR)
-	  continue;
-	fail("fcntl(F_DUPFD)");
-#endif /* FAKEROOT_FAKENET */
-      }
-#ifndef FAKEROOT_FAKENET
-    } else {
-#else /* FAKEROOT_FAKENET */
-    } while (0);
-
-    close(comm_sd);
-    comm_sd = fd;
-  }
-
-  val = 1;
-  if (setsockopt(comm_sd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof (val)) < 0)
-    fail("setsockopt(SO_REUSEADDR)");
+  if (fcntl(comm_sd, F_SETFD, FD_CLOEXEC) < 0)
+    fail("fcntl(F_SETFD, FD_CLOEXEC)");
 
   if (connect(comm_sd, get_addr(), sizeof (struct sockaddr_in)) < 0)
     fail("connect");
 }
 
-void close_comm_sd(void)
+void lock_comm_sd(void)
 {
-  /* We must synchronize since the vfork calls us in the parent process. */
   pthread_mutex_lock(&comm_sd_mutex);
+}
 
-  if (comm_sd >= 0) {
-    close(comm_sd);
-    comm_sd = -1;
-  }
-
+void unlock_comm_sd(void)
+{
   pthread_mutex_unlock(&comm_sd_mutex);
 }
 
+#endif /* FAKEROOT_FAKENET */
+
+#ifndef FAKEROOT_FAKENET
+
+void send_fakem(const struct fake_msg *buf)
+{
+  int r;
+
+  if(init_get_msg()!=-1){
+    ((struct fake_msg *)buf)->mtype=1;
+    r=msgsnd(msg_snd, (struct fake_msg *)buf,
+	     sizeof(*buf)-sizeof(buf->mtype), 0);
+    if(r==-1)
+      perror("libfakeroot, when sending message");
+  }
+}
+
+void send_get_fakem(struct fake_msg *buf)
+{
+  /*
+  send and get a struct fakestat from the daemon.
+  We have to use serial/pid numbers in addidtion to
+     the semaphore locking, to prevent the following:
+
+  Client 1 locks and sends a stat() request to deamon.
+  meantime, client 2 tries to up the semaphore too, but blocks.
+  While client 1 is waiting, it recieves a KILL signal, and dies.
+  SysV semaphores can eighter be automatically cleaned up when
+  a client dies, or they can stay in place. We have to use the
+  cleanup version, as otherwise client 2 will block forever.
+  So, the semaphore is cleaned up when client 1 recieves the KILL signal.
+  Now, client 1 falls through the semaphore_up, and
+  sends a stat() request to the daemon --  it will now recieve
+  the answer intended for client 1, and hell breaks lose (yes,
+  this has actually happened, and yes, it was hell (to debug)).
+
+  I realise that I may well do away with the semaphore stuff,
+  if I put the serial/pid numbers in the mtype field. But I cannot
+  store both PID and serial in mtype (just 32 bits on Linux). So
+  there will always be some (small) chance it will go wrong.
+  */
+
+  int l;
+  pid_t pid;
+  static int serial=0;
+
+  if(init_get_msg()!=-1){
+    pid=getpid();
+    serial++;
+    buf->serial=serial;
+    semaphore_up();
+    buf->pid=pid;
+    send_fakem(buf);
+
+    do
+      l=msgrcv(msg_get,
+               (struct my_msgbuf*)buf,
+               sizeof(*buf)-sizeof(buf->mtype),0,0);
+    while((buf->serial!=serial)||buf->pid!=pid);
+
+    semaphore_down();
+
+    /*
+    (nah, may be wrong, due to allignment)
+
+    if(l!=sizeof(*buf)-sizeof(buf->mtype))
+    printf("libfakeroot/fakeroot, internal bug!! get_fake: length=%i != l=%i",
+    sizeof(*buf)-sizeof(buf->mtype),l);
+    */
+
+  }
+}
+
+#else /* FAKEROOT_FAKENET */
+
+
+static size_t write_all(int fd,const void*buf,size_t count) {
+  ssize_t rc,remaining=count;
+  while(remaining>0) {
+	  rc= write(fd, buf+(count-remaining), remaining);
+	  if(rc<=0) {
+		  if(remaining==count) return rc;
+		  else fail("partial write");
+	  } else {
+		  remaining-=rc;
+	  }
+  }
+  return count-remaining;
+}
+
+static size_t read_all(int fd,void *buf,size_t count) {
+  ssize_t rc,remaining=count;
+  while(remaining>0) {
+	  rc = read(fd,buf+(count-remaining),remaining);
+	  if(rc<=0) {
+		  if(remaining==count) return rc;
+		  else fail("partial read");
+	  } else {
+		  remaining-=rc;
+	  }  
+  }
+  return count-remaining;
+}
+
 static void send_fakem_nr(const struct fake_msg *buf)
 {
   struct fake_msg fm;
@@ -374,115 +594,38 @@ static void send_fakem_nr(const struct fake_msg *buf)
   while (1) {
     ssize_t len;
 
-    len = write(comm_sd, &fm, sizeof (fm));
+    len = write_all(comm_sd, &fm, sizeof (fm));
     if (len > 0)
-#endif /* FAKEROOT_FAKENET */
       break;
-#ifdef FAKEROOT_FAKENET
 
     if (len == 0) {
       errno = 0;
       fail("write: socket is closed");
-#endif /* FAKEROOT_FAKENET */
     }
-#ifdef FAKEROOT_FAKENET
 
     if (errno == EINTR)
       continue;
 
     fail("write");
-#endif /* FAKEROOT_FAKENET */
   }
 }
 
 void send_fakem(const struct fake_msg *buf)
 {
-#ifndef FAKEROOT_FAKENET
-  int r;
-  
-  if(init_get_msg()!=-1){
-    ((struct fake_msg *)buf)->mtype=1;
-    r=msgsnd(msg_snd, (struct fake_msg *)buf, 
-	     sizeof(*buf)-sizeof(buf->mtype), 0);
-    if(r==-1)
-      perror("libfakeroot, when sending message");         
-  }
-#else /* FAKEROOT_FAKENET */
-  pthread_mutex_lock(&comm_sd_mutex);
+  lock_comm_sd();
 
   open_comm_sd();
   send_fakem_nr(buf);
 
-  pthread_mutex_unlock(&comm_sd_mutex);
-#endif /* FAKEROOT_FAKENET */
+  unlock_comm_sd();
 }
 
-#ifndef FAKEROOT_FAKENET
-void send_get_fakem(struct fake_msg *buf)
-{
-  /* 
-     send and get a struct fakestat from the daemon. 
-     We have to use serial/pid numbers in addidtion to 
-     the semaphore locking, to prevent the following:
-
-     Client 1 locks and sends a stat() request to deamon.
-     meantime, client 2 tries to up the semaphore too, but blocks.
-     While client 1 is waiting, it recieves a KILL signal, and dies.
-     SysV semaphores can eighter be automatically cleaned up when
-     a client dies, or they can stay in place. We have to use the
-     cleanup version, as otherwise client 2 will block forever.
-     So, the semaphore is cleaned up when client 1 recieves the KILL signal.
-     Now, client 1 falls through the semaphore_up, and
-     sends a stat() request to the daemon --  it will now recieve 
-     the answer intended for client 1, and hell breaks lose (yes,
-     this has actually happened, and yes, it was hell (to debug)).
-     
-     I realise that I may well do away with the semaphore stuff,
-     if I put the serial/pid numbers in the mtype field. But I cannot
-     store both PID and serial in mtype (just 32 bits on Linux). So 
-     there will always be some (small) chance it will go wrong.
-  */
-
-  int l;
-  pid_t pid;
-  static int serial=0;
-  
-#else /* FAKEROOT_FAKENET */
 static void get_fakem_nr(struct fake_msg *buf)
 {
   while (1) {
     ssize_t len;
-#endif /* FAKEROOT_FAKENET */
 
-#ifndef FAKEROOT_FAKENET
-  if(init_get_msg()!=-1){
-    pid=getpid();
-    serial++;
-    buf->serial=serial;
-    semaphore_up();
-    buf->pid=pid;
-    send_fakem(buf);
-    
-    do
-      l=msgrcv(msg_get,
-	       (struct my_msgbuf*)buf, 
-	       sizeof(*buf)-sizeof(buf->mtype),0,0);
-    while((buf->serial!=serial)||buf->pid!=pid);
-  
-    semaphore_down();
-  
-    /*
-      (nah, may be wrong, due to allignment)
-      
-      if(l!=sizeof(*buf)-sizeof(buf->mtype))
-      printf("libfakeroot/fakeroot, internal bug!! get_fake: length=%i != l=%i",
-      sizeof(*buf)-sizeof(buf->mtype),l);
-    */
- 
-  }
-  }
-#else /* FAKEROOT_FAKENET */
-    len = read(comm_sd, buf, sizeof (struct fake_msg));
+    len = read_all(comm_sd, buf, sizeof (struct fake_msg));
     if (len > 0)
       break;
     if (len == 0) {
@@ -507,93 +650,120 @@ static void get_fakem_nr(struct fake_msg *buf)
 
 void send_get_fakem(struct fake_msg *buf)
 {
-  pthread_mutex_lock(&comm_sd_mutex);
+  lock_comm_sd();
 
   open_comm_sd();
   send_fakem_nr(buf);
   get_fakem_nr(buf);
 
-  pthread_mutex_unlock(&comm_sd_mutex);
+  unlock_comm_sd();
 }
-#endif /* FAKEROOT_FAKENET */
-void send_stat(const struct stat *st,
-	       func_id_t f){
-  struct fake_msg buf;
-#ifndef FAKEROOT_FAKENET
-  
-  if(init_get_msg()!=-1){
-#else /* FAKEROOT_FAKENET */
 
 #endif /* FAKEROOT_FAKENET */
+
+void send_stat(const struct stat *st,
+	       func_id_t f
+#ifdef STUPID_ALPHA_HACK
+	       , int ver
+#endif
+	       ){
+  struct fake_msg buf;
+
+#ifndef FAKEROOT_FAKENET
+  if(init_get_msg()!=-1)
+#endif /* ! FAKEROOT_FAKENET */
+  {
+#ifndef STUPID_ALPHA_HACK
     cpyfakemstat(&buf,st);
+#else
+    cpyfakemstat(&buf,st,ver);
+#endif
     buf.id=f;
     send_fakem(&buf);
-#ifndef FAKEROOT_FAKENET
   }
-#endif /* ! FAKEROOT_FAKENET */
 }
 
 #ifdef STAT64_SUPPORT
 void send_stat64(const struct stat64 *st,
-#ifndef FAKEROOT_FAKENET
-	       func_id_t f){
-#else /* FAKEROOT_FAKENET */
-		 func_id_t f){
-#endif /* FAKEROOT_FAKENET */
+                 func_id_t f
+#ifdef STUPID_ALPHA_HACK
+                 , int ver
+#endif
+                 ){
   struct fake_msg buf;
-  
+
 #ifndef FAKEROOT_FAKENET
-  if(init_get_msg()!=-1){
+  if(init_get_msg()!=-1)
 #endif /* ! FAKEROOT_FAKENET */
+  {
+#ifndef STUPID_ALPHA_HACK
     cpyfakemstat64(&buf,st);
+#else
+    cpyfakemstat64(&buf,st,ver);
+#endif
     buf.id=f;
     send_fakem(&buf);
-#ifndef FAKEROOT_FAKENET
   }
-#endif /* ! FAKEROOT_FAKENET */
 }
 #endif /* STAT64_SUPPORT */
 
-void send_get_stat(struct stat *st){
+void send_get_stat(struct stat *st
+#ifdef STUPID_ALPHA_HACK
+		, int ver
+#endif
+		){
   struct fake_msg buf;
 
 #ifndef FAKEROOT_FAKENET
-  if(init_get_msg()!=-1){
+  if(init_get_msg()!=-1)
 #endif /* ! FAKEROOT_FAKENET */
+  {
+#ifndef STUPID_ALPHA_HACK
     cpyfakemstat(&buf,st);
-#ifndef FAKEROOT_FAKENET
-    
-#else /* FAKEROOT_FAKENET */
+#else
+    cpyfakemstat(&buf,st,ver);
+#endif
 
-#endif /* FAKEROOT_FAKENET */
     buf.id=stat_func;
     send_get_fakem(&buf);
+#ifndef STUPID_ALPHA_HACK
     cpystatfakem(st,&buf);
-#ifndef FAKEROOT_FAKENET
+#else
+    cpystatfakem(st,&buf,ver);
+#endif
   }
-#endif /* ! FAKEROOT_FAKENET */
 }
 
 #ifdef STAT64_SUPPORT
-
-void send_get_stat64(struct stat64 *st)
+void send_get_stat64(struct stat64 *st
+#ifdef STUPID_ALPHA_HACK
+                     , int ver
+#endif
+                    )
 {
   struct fake_msg buf;
 
 #ifndef FAKEROOT_FAKENET
-  if(init_get_msg()!=-1){
+  if(init_get_msg()!=-1)
 #endif /* ! FAKEROOT_FAKENET */
+  {
+#ifndef STUPID_ALPHA_HACK
     cpyfakemstat64(&buf,st);
-    
+#else
+    cpyfakemstat64(&buf,st,ver);
+#endif
+
     buf.id=stat_func;
     send_get_fakem(&buf);
+#ifndef STUPID_ALPHA_HACK
     cpystat64fakem(st,&buf);
-#ifndef FAKEROOT_FAKENET
+#else
+    cpystat64fakem(st,&buf,ver);
+#endif
   }
-#endif /* ! FAKEROOT_FAKENET */
 }
-
 #endif /* STAT64_SUPPORT */
+
 #ifndef FAKEROOT_FAKENET
 
 key_t get_ipc_key()
