@@ -19,6 +19,8 @@
 */
 #define _GNU_SOURCE 
 
+#define FAKEROOT_LIBFAKEROOT
+
 #include "config.h"
 #include "communicate.h"
 #include <stdlib.h>
@@ -33,7 +35,6 @@
 #include <dirent.h>
 #include <errno.h>
 
-
 /* 
    Where are those shared libraries? 
    If I knew of a configure/libtool way to find that out, I'd use it. Or
@@ -480,7 +481,11 @@ int WRAP_LSTAT LSTAT_ARG(int ver,
   r=NEXT_LSTAT(ver, file_name, statbuf);
   if(r)
     return -1;
+#ifndef STUPID_ALPHA_HACK
   send_get_stat(statbuf);
+#else
+  send_get_stat(statbuf, ver);
+#endif
   return 0;
 }
 
@@ -493,7 +498,11 @@ int WRAP_STAT STAT_ARG(int ver,
   r=NEXT_STAT(ver, file_name, st);
   if(r)
     return -1;
+#ifndef STUPID_ALPHA_HACK
   send_get_stat(st);
+#else
+  send_get_stat(st,ver);
+#endif
   return 0;
 }
 
@@ -508,7 +517,11 @@ int WRAP_FSTAT FSTAT_ARG(int ver,
   r=NEXT_FSTAT(ver, fd, st);
   if(r)
     return -1;
+#ifndef STUPID_ALPHA_HACK
   send_get_stat(st);
+#else
+  send_get_stat(st,ver);
+#endif
   return 0;
 }
 
@@ -526,8 +539,11 @@ int WRAP_LSTAT64 LSTAT64_ARG (int ver,
   if(r)
     return -1;
 
+#ifndef STUPID_ALPHA_HACK
   send_get_stat64(st);
-
+#else
+  send_get_stat64(st,ver);
+#endif
   return 0;
 }
 
@@ -540,7 +556,11 @@ int WRAP_STAT64 STAT64_ARG(int ver,
   r=NEXT_STAT64(ver,file_name,st);
   if(r)
     return -1;
+#ifndef STUPID_ALPHA_HACK
   send_get_stat64(st);
+#else
+  send_get_stat64(st,ver);
+#endif
   return 0;
 }
 
@@ -553,7 +573,11 @@ int WRAP_FSTAT64 FSTAT64_ARG(int ver,
   r=NEXT_FSTAT64(ver, fd, st);
   if(r)
     return -1;
+#ifndef STUPID_ALPHA_HACK
   send_get_stat64(st);
+#else
+  send_get_stat64(st,ver);
+#endif
 
   return 0;
 }
@@ -608,7 +632,11 @@ int chown(const char *path, uid_t owner, gid_t group){
     return r;
   st.st_uid=owner;
   st.st_gid=group;
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st,chown_func);
+#else
+  send_stat(&st,chown_func, _STAT_VER);
+#endif
   if(!dont_try_chown())
     r=next_lchown(path,owner,group);
   else
@@ -630,7 +658,11 @@ int lchown(const char *path, uid_t owner, gid_t group){
     return r;
   st.st_uid=owner;
   st.st_gid=group;
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st,chown_func);
+#else
+  send_stat(&st,chown_func, _STAT_VER);
+#endif
   if(!dont_try_chown())
     r=next_lchown(path,owner,group);
   else
@@ -652,7 +684,11 @@ int fchown(int fd, uid_t owner, gid_t group){
   
   st.st_uid=owner;
   st.st_gid=group;
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st, chown_func);
+#else
+  send_stat(&st, chown_func, _STAT_VER);
+#endif
   
   if(!dont_try_chown())
     r=next_fchown(fd,owner,group);
@@ -675,7 +711,11 @@ int chmod(const char *path, mode_t mode){
   
   st.st_mode=(mode&ALLPERMS)|(st.st_mode&~ALLPERMS);
     
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st, chmod_func);
+#else
+  send_stat(&st, chmod_func, _STAT_VER);
+#endif
   
   /* if a file is unwritable, then root can still write to it
      (no matter who owns the file). If we are fakeroot, the only
@@ -706,7 +746,11 @@ int fchmod(int fd, mode_t mode){
     return(r);
   
   st.st_mode=(mode&ALLPERMS)|(st.st_mode&~ALLPERMS);
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st,chmod_func);  
+#else
+  send_stat(&st,chmod_func, _STAT_VER);  
+#endif
   
   /* see chmod() for comment */
   mode |= 0600;
@@ -749,7 +793,11 @@ int WRAP_MKNOD MKNOD_ARG(int ver UNUSED,
   st.st_mode= mode & ~old_mask;
   st.st_rdev= XMKNOD_FRTH_ARG dev;
   
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st,mknod_func);
+#else
+  send_stat(&st,mknod_func, _STAT_VER);
+#endif
     
   return 0;
 }
@@ -777,7 +825,11 @@ int mkdir(const char *path, mode_t mode){
   
   st.st_mode=(mode&~old_mask&ALLPERMS)|(st.st_mode&~ALLPERMS)|S_IFDIR;
 
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st, chmod_func);
+#else
+  send_stat(&st, chmod_func, _STAT_VER);
+#endif
 
   return 0;
 }
@@ -815,7 +867,11 @@ int unlink(const char *pathname){
     return -1;
   
 #ifdef STAT64_SUPPORT
+#ifndef STUPID_ALPHA_HACK
   send_stat64(&st, unlink_func);
+#else
+  send_stat64(&st, unlink_func, _STAT_VER);
+#endif
 #else
   send_stat(&st, unlink_func);
 #endif
@@ -838,7 +894,11 @@ int rmdir(const char *pathname){
   if(r)
     return -1;
 
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st,unlink_func);  
+#else
+  send_stat(&st,unlink_func, _STAT_VER);  
+#endif
 
   return 0;
 }
@@ -857,7 +917,11 @@ int remove(const char *pathname){
   r=next_remove(pathname);  
   if(r)
     return -1;
+#ifndef STUPID_ALPHA_HACK
   send_stat(&st,unlink_func);  
+#else
+  send_stat(&st,unlink_func, _STAT_VER);  
+#endif
   
   return r;
 }
@@ -887,7 +951,11 @@ int rename(const char *oldpath, const char *newpath){
   if(s)
     return -1;
   if(!r)
+#ifndef STUPID_ALPHA_HACK
     send_stat(&st,unlink_func);
+#else
+    send_stat(&st,unlink_func, _STAT_VER);
+#endif
 
   return 0;
 }
@@ -899,8 +967,13 @@ pid_t fork(void)
 
   pid = next_fork();
 
-  if (pid == 0)
-    close_comm_sd();
+  if (pid == 0) {
+    /* No need to lock in the child process. */
+    if (comm_sd >= 0) {
+      next_close(comm_sd);
+      comm_sd = -1;
+    }
+  }
 
   return pid;
 }
@@ -911,7 +984,50 @@ pid_t vfork(void)
   return fork();
 }
 
+/* Return an error when trying to close the comm_sd file descriptor
+   (pretend that it's closed). */
+int close(int fd)
+{
+  int retval, reterr;
+
+  lock_comm_sd();
+
+  if (comm_sd >= 0 && comm_sd == fd) {
+    retval = -1;
+    reterr = EBADF;
+  } else {
+    retval = next_close(fd);
+    reterr = errno;
+  }
+
+  unlock_comm_sd();
+
+  errno = reterr;
+  return retval;
+}
+
+int dup2(int oldfd, int newfd)
+{
+  int retval, reterr;
+
+  lock_comm_sd();
+
+  if (comm_sd >= 0 && comm_sd == newfd) {
+    /* If dup fails, comm_sd gets set to -1, which is fine. */
+    comm_sd = dup(newfd);
+    next_close(newfd);
+  }
+
+  retval = next_dup2(oldfd, newfd);
+  reterr = errno;
+
+  unlock_comm_sd();
+
+  errno = reterr;
+  return retval;
+}
 #endif /* FAKEROOT_FAKENET */
+
 uid_t getuid(void){
   if (fakeroot_disabled)
     return next_getuid();
